# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio
import os
os.environ['PYTHON_SETTINGS_MODULE'] = __name__

from cbra.ext.sendgrid import SendgridEmailSender


EMAIL_SERVER: str = 'sendgrid'

EMAIL_CREDENTIAL: str = os.environ['SENDGRID_API_KEY']


async def main():
    sender = SendgridEmailSender()
    await sender.send_template(
        sender="Molano <noreply@suppliers.molanohq.com>",
        recipients=["cochise.ruhulessin@unimatrixone.io"],
        subject="Please confirm order by Molano",
        context={
            'access_code': '012345',
            'purchase_order_url': 'https://suppliers.molanohq.com/orders/1'
        },
        templates={
            'text/plain': 'notification.txt'
        }   
    )


if __name__ == '__main__':
    asyncio.run(main())


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic
from canonical import EmailAddress

from ..types import PairwiseIdentifier
from ..types import ResourceOwnerIdentifier
from ..types import RequestedScope


class ResourceOwner(pydantic.BaseModel):
    client_id: str
    consents: set[str] = set()
    email: EmailAddress | None = None
    ppid: PairwiseIdentifier

    @property
    def id(self) -> ResourceOwnerIdentifier:
        return ResourceOwnerIdentifier(client_id=self.client_id, sub=self.ppid.sub)
    
    @property
    def sub(self) -> int:
        return self.ppid.sub
    
    def consent_required(self, scope: list[RequestedScope]) -> list[RequestedScope]:
        return [
            x for x in scope
            if x.name not in self.consents
            and x.requires_consent()
        ]


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio
from typing import Any
from typing import Callable


class Runner:

    @staticmethod
    async def run_in_executor(
        func: Callable[..., Any],
        *args: Any,
        **kwargs: Any
    ) -> Any:
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, func, *args, **kwargs)


# Copyright (C) 2020-2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pytest

import aorta
from cbra.ext.google import GooglePubsubTransport


class GoogleEvent(aorta.Event):
    pass


class GoogleCommand(aorta.Command):
    pass


@pytest.fixture
def transport() -> GooglePubsubTransport:
    return GooglePubsubTransport(
        project='project',
        prefix='prefix',
        service_name='service_name'
    )


def test_topics_event(
    transport: GooglePubsubTransport
):
    envelope = GoogleEvent().envelope()
    topics = set(transport.get_topics(envelope))
    assert topics == {
        "projects/project/topics/prefix.events",
        "projects/project/topics/prefix.events.GoogleEvent"
    }


def test_topics_command(
    transport: GooglePubsubTransport
):
    envelope = GoogleCommand().envelope()
    topics = set(transport.get_topics(envelope))
    assert topics == {
        "projects/project/topics/prefix.commands.service_name",
    }


def test_topics_command_audience(
    transport: GooglePubsubTransport
):
    envelope = GoogleCommand().envelope()
    envelope.metadata.audience = ['foo']
    topics = set(transport.get_topics(envelope))
    assert topics == {
        "projects/project/topics/prefix.commands.foo",
    }


def test_topics_command_audiences(
    transport: GooglePubsubTransport
):
    envelope = GoogleCommand().envelope()
    envelope.metadata.audience = ['foo', 'bar', 'baz']
    topics = set(transport.get_topics(envelope))
    assert topics == {
        "projects/project/topics/prefix.commands.foo",
        "projects/project/topics/prefix.commands.bar",
        "projects/project/topics/prefix.commands.baz",
    }


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from ckms.core import Keychain
from ckms.types import JSONWebKeySet

from cbra.core.params import ApplicationKeychain
from .endpoint import AuthorizationServerEndpoint


class ClientKeysEndpoint(AuthorizationServerEndpoint):
    keychain: Keychain = ApplicationKeychain
    name: str = 'oauth2.client-jwks'
    status_code: int = 303
    summary: str = 'Server JWKS Endpoint'
    path: str = '/client-jwks.json'

    async def get(self) -> JSONWebKeySet:
        return self.keychain.as_jwks(private=False)


# Copyright (C) 2020-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from cbra.core import Endpoint

from .debugcommand import DebugCommand
from .debugevent import DebugEvent


class AortaDebugEndpoint(Endpoint):
    __module__: str = 'cbra.ext.google'
    include_in_schema: bool = False

    async def post(self) -> None:
        self.transaction.publish(DebugCommand())
        self.transaction.publish(DebugCommand(), audience={'self'})
        self.transaction.publish(DebugEvent())


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio

import aorta
import pytest

import cbra.core as cbra


@pytest.fixture(scope="session")
def event_loop():
    loop = asyncio.new_event_loop()
    yield loop
    loop.close()


@pytest.fixture
def publisher() -> cbra.MessagePublisher:
    return cbra.MessagePublisher(
        transport=aorta.NullTransport(),
        request=None # type: ignore
    )


@pytest.fixture
def runner(
    publisher: cbra.MessagePublisher
) -> aorta.types.IRunner:
    return cbra.MessageRunner(
        publisher=publisher,
        request=None # type: ignore
    )


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import urllib.parse

import pytest

from cbra.ext import oauth2
from .conftest import ServerClient
from .conftest import SUPPORTED_RESPONSE_TYPES


@pytest.mark.parametrize("response_type", SUPPORTED_RESPONSE_TYPES)
@pytest.mark.parametrize("state", [None, "state"])
@pytest.mark.asyncio
async def test_with_default_redirect_uri(
    client: ServerClient,
    authorization_endpoint: str,
    response_type: oauth2.types.ResponseType,
    state: str | None,
    issuer: str
):
    await client.login()
    params: dict[str, str] = {
        'client_id': 'app',
        'response_type': response_type.value
    }
    if state is not None:
        params['state'] = state
    response = await client.get(
        url=authorization_endpoint,
        params=params
    )
    assert response.status_code == 303, response.content
    if response_type == oauth2.types.ResponseType.code:
        assert 'Location' in response.headers
        p = urllib.parse.urlparse(response.headers['Location'])
        q = dict(urllib.parse.parse_qsl(p.query))
        assert 'code' in q
        assert 'iss' in q
        assert 'error' not in q
        assert q['iss'] == issuer
        if state is not None:
            assert 'state' in q
            assert q['state'] == state


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import functools
import warnings
from typing import Any

from headless.ext.oauth2 import Client

from cbra.types import HMACSignature
from cbra.types import ICredential
from cbra.types import ICredentialVerifier
from cbra.types import IDependant
from cbra.types import OIDCRequestPrincipal
from cbra.types import RFC9068RequestPrincipal
from cbra.types import SessionRequestPrincipal
from cbra.types import JSONWebToken
from ..ioc.config import TRUSTED_AUTHORIZATION_SERVERS
from ..params import ApplicationSecretKey
from ..secretkey import SecretKey


class AuthenticationService(
    ICredentialVerifier[RFC9068RequestPrincipal|OIDCRequestPrincipal|SessionRequestPrincipal],
    IDependant
):
    __module__: str = 'cbra.core.iam'
    providers: set[str]

    def __init__(
        self,
        providers: list[str] = TRUSTED_AUTHORIZATION_SERVERS,
        secret_key: SecretKey = ApplicationSecretKey
    ):
        self.providers = set(providers)
        self.secret_key = secret_key

    @functools.singledispatchmethod # type: ignore
    async def verify(
        self,
        principal: OIDCRequestPrincipal | RFC9068RequestPrincipal | SessionRequestPrincipal,
        credential: ICredential | None,
        providers: set[str] | None = None
    ) -> bool:
        if providers: raise NotImplementedError
        warnings.warn(
            f'Unknown principal {type(principal).__name__}. '
            f'{type(self).__name__}.verify(principal, credential) '
            'will always return False.'
        )
        return False

    @verify.register
    async def verify_oidc(
        self,
        principal: OIDCRequestPrincipal,
        credential: JSONWebToken,
        providers: set[str] | None = None
    ) -> bool:
        return await self.verify_oauth_jwt(principal, credential, providers)

    async def verify_oauth_jwt(
        self,
        principal: OIDCRequestPrincipal | RFC9068RequestPrincipal,
        credential: JSONWebToken,
        providers: set[str] | None = None
    ) -> bool:
        providers = providers or self.providers
        if principal.iss not in providers:
            return False
        async with Client(issuer=principal.iss) as client:
            return await client.verify(str(credential))

    @verify.register
    async def verify_session(
        self,
        principal: SessionRequestPrincipal,
        credential: HMACSignature,
        *args: Any,
        **kwargs: Any
    ) -> bool:
        return await credential.verify(self.secret_key)


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
import ipaddress

from .isubject import ISubject


class IAuthorizationContext:
    __module__: str = 'cbra.types'
    timestamp: datetime.datetime

    @property
    def remote_host(self) -> ipaddress.IPv4Address | None:
        """A :class:`~ipaddress.IPv4Address` instance specifying the remote
        host from which a request was made, or ``None`` if there is no
        network request.
        """
        return self.get_remote_host()

    @property
    def subject(self) -> ISubject:
        """The subject that was resolved from the principal and the
        credential.
        """
        return self.get_subject()

    async def authorize(self) -> bool:
        """Fetch the permissions, roles and other authorizations for the current
        subject. When implementing :class:`IAuthorizationContext`, provide an
        awaitable object to the constructor that knows how to retrieve the
        authorizations.
        """
        raise NotImplementedError

    def is_authenticated(self) -> bool:
        """Return a boolean indicating if the request is authenticated."""
        return self.subject.is_authenticated()

    def get_remote_host(self) -> ipaddress.IPv4Address | None:
        raise NotImplementedError

    def get_subject(self) -> ISubject:
        raise NotImplementedError

    def has_permission(self, name: str) -> bool:
        """Return a boolean indicating if the context has the given
        permission. This method should always be invoked after
        :meth:`authorize()`.
        """
        raise NotImplementedError


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import TypeVar

import fastapi
import pydantic

from cbra.types import IEndpoint
from .iresource import IResource
from .resourceaction import ResourceAction


T = TypeVar('T', bound='CreateAction')


class CreateAction(ResourceAction):
    action: str = 'create'
    name_template: str = 'Create a new {name}'
    status_code: int = 201

    @classmethod
    def fromfunc(
        cls: type[T],
        name: str,
        func: Any
    ) -> T:
        return cls(name=name, method='POST', func=func)

    def can_write(self) -> bool:
        return True

    def get_return_annotation(self) -> Any:
        return self.endpoint.model.__response_model__

    def get_openapi_responses(
        self,
        cls: type[IResource],
        responses: dict[int | str, Any]
    ) -> dict[int | str, Any]:
        responses.update({
            409: {
                'description': (
                    'The unique properties of an existing '
                    f'**{self.endpoint.verbose_name}** conflict'
                )
            }
        })
        return super().get_openapi_responses(cls, responses)

    def get_write_model(self) -> type[pydantic.BaseModel]:
        return self.endpoint.model.__create_model__

    def is_detail(self) -> bool:
        return False

    def needs_resource(self) -> bool:
        return True

    def parse_resource(self, resource: pydantic.BaseModel) -> pydantic.BaseModel:
        return self.endpoint.model.__create_model__.parse_obj(resource.dict())

    async def process_response(
        self,
        endpoint: IEndpoint,
        response: fastapi.Response | pydantic.BaseModel | None
    ) -> fastapi.Response:
        if isinstance(response, pydantic.BaseModel):
            response = self.response_model.parse_obj(response.dict())
        return await super().process_response(endpoint, response)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

from cbra.ext.webhooks import WebhookEndpoint

from .shopifywebhookenvelope import ShopifyWebhookEnvelope


class ShopifyWebhookEndpoint(WebhookEndpoint):
    __module__: str = 'cbra.ext.shopify'
    domain: str = "shopify.com"
    envelope: ShopifyWebhookEnvelope
    signed: bool = True
    summary: str = "Shopify"
    tags: list[str] = ["Webhooks (Incoming)"]

    def log_event(self, dto: Any) -> None:
        self.logger.info("Received event %s", self.envelope.event_name)

    async def on_orders_cancelled(
        self,
        dto: Any
    ) -> None:
        self.log_event(dto)

    async def on_orders_create(
        self,
        dto: Any
    ) -> None:
        self.log_event(dto)

    async def on_orders_delete(
        self,
        dto: Any
    ) -> None:
        self.log_event(dto)

    async def on_orders_edited(
        self,
        dto: Any
    ) -> None:
        self.log_event(dto)

    async def on_orders_fulfilled(
        self,
        dto: Any
    ) -> None:
        self.log_event(dto)

    async def on_orders_paid(
        self,
        dto: Any
    ) -> None:
        self.log_event(dto)

    async def on_orders_partially_fulfilled(
        self,
        dto: Any
    ) -> None:
        self.log_event(dto)

    async def on_orders_updated(
        self,
        dto: Any
    ) -> None:
        self.log_event(dto)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime

import pydantic


class Event(pydantic.BaseModel):
    idhook: int = pydantic.Field(
        default=...,
        title="Webhook ID",
        description=(
            'Identifies the webhook that produced the incoming '
            'message.'
        )
    )

    name: str = pydantic.Field(
        default=...,
        title='Name',
        description=(
            'The name of the webhook as created in the Picqer '
            'application.'
        )
    )

    event_triggered_at: datetime.datetime = pydantic.Field(
        default=...,
        title="Timestamp",
        description=(
            'The date and time at which the event was triggered.'
        )
    )

    event: str


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import pydantic

from cbra.types import PolicyPrincipal


class AccessControlList(pydantic.BaseModel):
    enabled: bool = True
    members: list[PolicyPrincipal] = []

    def is_enabled(self) -> bool:
        """Return a boolean indicating if the ACL is enabled."""
        return self.enabled

    def is_member(self, obj: Any) -> bool:
        """Return a boolean indicating if the object is a member of the
        ACL.
        """
        return any([x.match(obj) for x in self.members])



# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pytest
from headless.core import httpx


@pytest.mark.asyncio
async def test_missing_client_id_returns_400(
    client: httpx.Client,
    authorization_endpoint: str
):
    response = await client.get(
        url=authorization_endpoint,
        params={}
    )
    assert response.status_code == 400, response.content
    assert 'Content-Type' in response.headers
    assert response.headers['Content-Type'] == 'text/plain; charset=utf-8'


@pytest.mark.asyncio
async def test_invalid_client_id_returns_400(
    client: httpx.Client,
    authorization_endpoint: str,
    random_id: str
):
    response = await client.get(
        url=authorization_endpoint,
        params={'client_id': random_id}
    )
    assert response.status_code == 400, response.content
    assert 'Content-Type' in response.headers
    assert response.headers['Content-Type'] == 'text/plain; charset=utf-8'


# Copyright (C) 2020-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""Declares :class:`EventarcEndpoint`."""
import cbra.core as cbra

from .googleendpoint import GoogleEndpoint
from .messagepublished import MessagePublished
from .pubsubmessage import PubsubMessage


class EventarcEndpoint(GoogleEndpoint):
    """Accepts a message from Google Eventarc from a HTTP POST request. The
    default implementation does nothing. Implementations must override
    the :meth:`on_message()` method to handle the incoming messages.
    """
    __module__: str = 'cbra.ext.google'
    include_in_schema: bool = False
    status_code: int = 202
    summary: str = 'Eventarc Message'
    tags: list[str] = ['Cloud Endpoints']

    @cbra.describe(status_code=202)
    async def post(self, dto: MessagePublished) -> dict[str, bool]:
        """Receive a `google.cloud.pubsub.topic.v1.messagePublished` message
        and invoke the appropriate handler.
        """
        self.logger.debug(
            "Received message %s from Google Pub/Sub",
            dto.message.message_id
        )
        await self.on_message(dto.message)
        return {'success': True}

    async def on_message(self, message: PubsubMessage) -> None:
        """Handles the message received from Google Pub/Sub. The default
        implementation does nothing
        """
        return


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import enum


class ClientAuthenticationMethod(str, enum.Enum):
    none        = 'none'
    post        = 'client_secret_post'
    basic       = 'client_secret_basic'
    jwt         = 'client_secret_jwt'
    private_key = 'private_key_jwt'
    tls         = 'tls_client_auth'
    self_signed = 'self_signed_tls_client_auth'


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import logging

import fastapi
import pydantic
from headless.ext.oauth2.types import GrantType

import cbra.core as cbra
from cbra.core.iam.types import Subject
from ..authorizationserverstorage import AuthorizationServerStorage
from ..models import AuthorizationRequest
from ..models import Client
from ..models import Grant
from ..models import ResourceOwner
from ..types import AuthorizationCode
from ..types import FatalClientException
from ..types import InvalidRequest
from .requestingclient import RequestingClient


logger: logging.Logger = logging.getLogger('cbra.endpoint')


async def get(
    storage: AuthorizationServerStorage = cbra.instance('_AuthorizationServerStorage'),
    client: Client = RequestingClient,
    grant_type: str = fastapi.Form(
        default=...,
        title="Grant type"
    ),
    code: AuthorizationCode | None = fastapi.Form(
        default=None,
        title="Code",
        description=(
            "The authorization code received from the authorization server. "
            "This parameter is **required** when `grant_type` is "
            "`authorization_code`, otherwise it is ignored."
        )
    ),
    redirect_uri: str | None = fastapi.Form(
        default=None,
        title="Redirect URI",
        description=(
            "This parameter is **required**, if the `redirect_uri` parameter "
            "was included in the authorization request, in which case their "
            "values **must"" be identical.  If no `redirect_uri` was included "
            "in the authorization request, this parameter is **optional**."
        )
    )
):
    try:
        grant_type = GrantType(grant_type)
    except ValueError:
        raise FatalClientException(
            code='unsupported_grant_type',
            message=(
                "The authorization server does not support the given grant:"
                f"{grant_type[:64]}"
            )
        )
    if not client.can_grant(grant_type):
        raise FatalClientException(
            code='unauthorized_client',
            message=(
                "The client does not allow or is not allowed to use the "
                f"given grant: {grant_type[:64]}"
            )
        )
    
    authnrequest: AuthorizationRequest | None = None
    owner: ResourceOwner | None = None
    subject: Subject | None = None
    if grant_type == GrantType.authorization_code:
        if code is None:
            raise InvalidRequest("The authorization code is not accepted by the server.")
        authnrequest = await storage.get(AuthorizationRequest, code)
    if authnrequest is not None:
        assert authnrequest.has_owner()
        owner = await storage.get(ResourceOwner, authnrequest.owner)
        subject = await storage.get(Subject, authnrequest.owner.sub)
    if owner is None:
        logger.critical("Resource owner not found (client: %s)", client.client_id)
    if subject is None:
        logger.critical("Subject not found (client: %s)", client.client_id)
    try:
        return Grant.parse_obj({
            'grant_type': grant_type,
            'code': code,
            'client': client,
            'owner': owner,
            'redirect_uri': redirect_uri,
            'request': authnrequest,
            'subject': subject
        })
    except pydantic.ValidationError as e:
        logger.debug("Caught fatal ValidationError: %s", repr(e))
        raise InvalidRequest(
            f"The parameters for grant {grant_type[:64]} were invalid, "
            f"incomplete, duplicated or otherwise malformed."
        )


RequestedGrant: Grant = fastapi.Depends(get)


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi
from ckms.core import Keychain
from ckms.core import parse_dsn
from ckms.core import parse_spec

from cbra.core.conf import settings


keychain: Keychain = Keychain()


async def get_application_keychain() -> Keychain:
    must_configure: bool = False
    if settings.APP_ENCRYPTION_KEY and not keychain.has('enc'):
        keychain.add('enc', parse_spec(parse_dsn(settings.APP_ENCRYPTION_KEY)))
        must_configure = True
    if settings.APP_SIGNING_KEY and not keychain.has('sig'):
        keychain.add('sig', parse_spec(parse_dsn(settings.APP_SIGNING_KEY)))
        must_configure = True
    if must_configure:
        await keychain
    return keychain


ApplicationKeychain: Keychain = fastapi.Depends(get_application_keychain)


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import os

import pytest
import pytest_asyncio
from headless.core import httpx

import cbra.core as cbra
from cbra.core.resource.test import Book
from cbra.core.resource.test import BookResource


@pytest_asyncio.fixture(scope='session') # type: ignore
async def client():
    app = cbra.Application()
    BookResource.add_to_router(app, path='/')
    async with httpx.Client(base_url='https://cbra', app=app) as client:
        yield client


@pytest_asyncio.fixture # type: ignore
async def book(client: httpx.Client):
    response = await client.post(
        url='/books',
        json={
            'title': f"Foo {os.urandom(10).hex()}",
            'publications': []
        }
    )
    assert response.status_code == 201
    resource = Book.parse_obj(await response.json())
    yield resource
    response = await client.delete(url=f'/books{resource.id}', allow_none=True)
    assert response.status_code in {200, 404}


@pytest.mark.asyncio
async def test_get(
    client: httpx.Client
):
    response = await client.get(url='/books/1')
    resource = Book.parse_obj(await response.json())
    assert resource.id == 1


@pytest.mark.asyncio
async def test_post(
    client: httpx.Client
):
    response = await client.post(
        url='/books',
        json={
            'title': "Foo",
            'publications': []
        }
    )
    resource = Book.parse_obj(await response.json())
    assert resource.id != 1


@pytest.mark.asyncio
async def test_delete(
    client: httpx.Client,
    book: Book
):
    response = await client.delete(url=f'/books/{book.id}')
    assert response.status_code == 200

    response = await client.get(url=f'/books/{book.id}', allow_none=True)
    assert response.status_code == 404


@pytest.mark.asyncio
async def test_update(
    client: httpx.Client,
    book: Book
):
    title = f'Foo {os.urandom(10).hex()}'
    response = await client.request(
        method='PATCH',
        url=f'/books/{book.id}',
        json={'title': title}
    )
    assert response.status_code == 200
    new = Book.parse_obj(await response.json())
    assert new.id == book.id
    assert new.title == title


@pytest.mark.asyncio
async def test_replace(
    client: httpx.Client,
    book: Book
):
    title = f'Foo {os.urandom(10).hex()}'

    # Must be full Book resource.
    response = await client.request(
        method='PUT',
        url=f'/books/{book.id}',
        json={'title': title}
    )
    assert response.status_code == 422

    response = await client.request(
        method='PUT',
        url=f'/books/{book.id}',
        json={**book.dict(), 'title': title}
    )
    assert response.status_code == 200
    new = Book.parse_obj(await response.json())
    assert new.id == book.id
    assert new.title == title


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import secrets
from datetime import datetime
from datetime import timezone

import pydantic

from ..types import AuthorizationLifecycle


class Authorization(pydantic.BaseModel):
    """Maintains the lifecycle of an OAuth 2.x/OpenID Connect
    authorization request.
    """
    request_id: str = pydantic.Field(
        default_factory=lambda: secrets.token_bytes(48)
    )

    created: datetime = pydantic.Field(
        default_factory=lambda: datetime.now(timezone.utc)
    )

    status: AuthorizationLifecycle = pydantic.Field(
        default=AuthorizationLifecycle.requested
    )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Generator
from typing import TypeVar


T = TypeVar('T', bound='IDeferred')


class IDeferred:
    __module__: str = 'cbra.types'

    @classmethod
    def defer(cls: type[T], obj: T, name: str) -> Any:
        return cls.DeferredAttribute(obj, name)

    async def initialize(self) -> None:
        raise NotImplementedError

    async def _initialize(self: T) -> T:
        await self.initialize()
        return self

    def __await__(self: T) -> Generator[None, None, T]:
        return self._initialize().__await__()

    class DeferredAttribute:

        def __init__(self, obj: 'IDeferred', name: str):
            self.obj = obj
            self.name = name

        def __getattr__(self, __name: str) -> Any:
            raise RuntimeError(
                f'Await {type(self.obj).__name__} before accessing attribute '
                f'{type(self.obj).__name__}.{self.name}.{__name}.'
            )

        __repr__ = lambda self: self.__getattr__('__repr__') # type: ignore


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import logging
from typing import cast
from typing import Any
from typing import Literal

import pydantic
from headless.ext.oauth2.models import TokenResponse
from headless.ext.oauth2.types import GrantType

from cbra.core.iam.models import Subject
from ..types import AuthorizationCode
from ..types import ITokenBuilder
from ..types import ITokenSigner
from .authorizationrequest import AuthorizationRequest
from .client import Client
from .resourceowner import ResourceOwner


logger: logging.Logger = logging.getLogger('cbra.endpoint')


class AuthorizationCodeGrant(pydantic.BaseModel):
    code: AuthorizationCode
    client: Client
    grant_type: Literal[GrantType.authorization_code]
    owner: ResourceOwner
    redirect_uri: str | None = None
    request: AuthorizationRequest
    subject: Subject

    @pydantic.validator('request')
    def validate_request(
        cls,
        request: AuthorizationRequest | None
    ) -> AuthorizationRequest:
        if request is None:
            raise ValueError('authorization request missing.')
        if not request.has_owner():
            raise ValueError('authorization request does not have an owner.')
        return request
    
    @pydantic.root_validator
    def validate_grant(
        cls,
        values: dict[str, Any]
    ) -> dict[str, Any]:
        client = cast(Client, values.get('client'))
        owner = cast(ResourceOwner, values.get('owner'))
        redirect_uri = cast(str, values.get('redirect_uri'))
        request = cast(AuthorizationRequest, values.get('request'))
        is_valid: bool = all([
            client,
            request,
            redirect_uri,
            owner
        ])
        if not is_valid:
            return values

        if not request.is_owned_by(owner.sub):
            raise ValueError('request was made by a different Subject.')

        if str(redirect_uri) != str(request.redirect_uri):
            logger.debug(
                "Redirect URI mismatch (expected: %s, actual: %s)",
                request.redirect_uri, redirect_uri
            )
            raise ValueError(
                'redirect URI does not correspond with the redirect URI '
                'requested.'
            )

        return values
    
    async def create_response(
        self,
        builder: ITokenBuilder,
        signer: ITokenSigner
    ) -> TokenResponse:
        assert self.owner.ppid.value is not None
        at, expires_in = builder.rfc9068(
            sub=self.owner.ppid.value,
            scope=self.request.scope,
            auth_time=self.request.auth_time
        )
        access_token = await at.sign(signer)
        id_token = None
        refresh_token = None
        if self.request.has_refresh_token():
            refresh_token = await builder.refresh_token(
                grant_type=GrantType.authorization_code,
                client_id=self.client.client_id,
                sector_identifier=self.client.sector_identifier,
                sub=self.owner.sub,
                ppid=self.owner.ppid.value,
                scope=self.request.scope,
                renew=self.client.refresh_token_policy,
                ttl=self.client.refresh_token_ttl
            )
        if self.request.has_id_token():
            assert self.request.nonce is not None
            id_token = builder.id_token(
                subject=self.subject,
                ppid=self.owner.ppid.value,
                access_token=access_token,
                nonce=self.request.nonce,
                scope=self.request.scope,
                auth_time=self.request.auth_time,
                authorization_code=self.request.code,
                request=self.request,
                owner=self.owner
            )
            id_token = await id_token.sign(signer)
        return TokenResponse.parse_obj({
            'token_type': 'Bearer',
            'access_token': access_token,
            'expires_in': expires_in,
            'id_token': id_token,
            'refresh_token': refresh_token,
            'scope': str.join(' ', sorted([x.name for x in self.request.scope]))
        })


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pytest

import google.oauth2.id_token
import google.auth.transport.requests
from google.auth.exceptions import DefaultCredentialsError
from google.cloud.datastore import Client
from cbra.ext.google.environ import GOOGLE_DATASTORE_NAMESPACE
from cbra.ext.google.environ import GOOGLE_SERVICE_PROJECT


@pytest.fixture(scope='session')
def google_id_token() -> str:
    audience = 'http://cbra.ext.google'
    try:
        request = google.auth.transport.requests.Request()
        return google.oauth2.id_token.fetch_id_token(request, audience) # type: ignore
    except DefaultCredentialsError:
        pytest.skip("Google default credentials not found")


@pytest.fixture(scope='session')
def client() -> Client:
    if not GOOGLE_SERVICE_PROJECT or not GOOGLE_DATASTORE_NAMESPACE:
        pytest.skip("Google not configured")
    return Client(
        project=GOOGLE_SERVICE_PROJECT,
        namespace=GOOGLE_DATASTORE_NAMESPACE
    )


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

from cbra.types import NotFound
from .detail import Detail


class Retrieve(Detail):
    __module__: str = 'cbra.core'

    async def retrieve(self, *args: Any) -> Any:
        resource = self.adapt(await self.get_object()) # type: ignore
        if resource is None:
            raise NotFound
        return resource


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import TypeVar

from .detailaction import DetailAction
from .iresource import IResource


T = TypeVar('T', bound='CustomAction')


class CustomAction(DetailAction):
    action: str
    status_code: int
    path: str
    use_resource_model: bool = False
    _summary: str

    @property
    def summary(self) -> str:
        return self._summary

    @classmethod
    def fromfunc(
        cls: type[T],
        name: str,
        func: Any
    ) -> T:
        if not hasattr(func, 'action'):
            raise TypeError(f"{name}.{func.__name__} is not an action.")
        return cls(name=name, func=func, **func.action)
    
    def __init__(
        self,
        action: str,
        status_code: int,
        summary: str,
        *args: Any,
        **kwargs: Any
    ):
        self.action = action
        self.path = action
        self.status_code = status_code
        self._summary = summary
        super().__init__(*args, **kwargs)

    def can_mutate(self) -> bool:
        # Custom actions are always detail actions, so POST
        # requests are considered mutations.
        return super().can_mutate() or self.method == "POST"

    def get_url_pattern(
        self,
        prefix: str | None,
        endpoint: type[IResource] | None = None
    ) -> str:
        return f'{super().get_url_pattern(prefix, endpoint)}/{self.path}'

    def is_detail(self) -> bool:
        return True

    def needs_resource(self) -> bool:
        return False


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import AsyncGenerator

import fastapi

import cbra.core as cbra
from ..models import Client
from ..types import FatalAuthorizationException
from ..types import IAuthorizationServerStorage
from ..types import OIDCProvider
from .cookieclientidentifier import CookieClientIdentifier
from .localclientprovider import ClientProvider
from .localclientprovider import LocalClientProvider


__all__: list[str] = ['LocalOpenIdProvider']


async def get(
    client_id: str | None = CookieClientIdentifier,
    clients: ClientProvider = LocalClientProvider,
    storage: IAuthorizationServerStorage = cbra.instance(
        name='_AuthorizationServerStorage'
    ),
) -> AsyncGenerator[OIDCProvider, None]:
    if client_id is None:
        raise FatalAuthorizationException(
            "The 'client_id' cookie is not provided."
        )
    client = await clients.get(client_id)
    if client is not None:
        provider = client
    else:
        client = await storage.get(Client, client_id)
        if client is None:
            raise FatalAuthorizationException(
                "The client specified by the request does not exist."
            )
        provider = client.get_provider()
    async with provider:
        yield provider


LocalOpenIdProvider: OIDCProvider = fastapi.Depends(get)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import functools
from typing import Any

from headless.ext.oauth2.models import TokenResponse

import cbra.core as cbra
from cbra.types import NullRequestPrincipal
from .endpoint import AuthorizationServerEndpoint
from .models import AuthorizationCodeGrant
from .models import Client
from .models import Grant
from .params import RequestingClient
from .params import RequestedGrant
from .params import RequestResourceOwner
from .params import TokenBuilder
from .params import TokenSigner
from .types import ITokenBuilder
from .types import ITokenSigner


class TokenEndpoint(AuthorizationServerEndpoint):
    __module__: str = 'cbra.ext.oauth2'
    builder: ITokenBuilder = TokenBuilder.depends()
    client: Client = RequestingClient # type: ignore
    metrics: cbra.MetricReporter = cbra.MetricReporter('oauth2')
    name: str = 'oauth2.token'
    summary: str = 'Token Endpoint'
    owner: RequestResourceOwner
    path: str = '/token'
    principal: NullRequestPrincipal = NullRequestPrincipal.depends()
    signer: ITokenSigner = TokenSigner.depends()

    async def post(self, grant: Grant = RequestedGrant) -> TokenResponse:
        return await grant.handle(self.handle_grant)

    @functools.singledispatchmethod
    async def handle_grant(
        self,
        grant: Any
    ) -> TokenResponse:
        raise NotImplementedError(type(grant).__name__)
    
    @handle_grant.register
    async def authorization_code(
        self,
        grant: AuthorizationCodeGrant
    ) -> TokenResponse:
        assert self.request.client is not None
        response = await grant.create_response(self.builder, self.signer)
        self.metrics.report('GrantAuthorized', {
            'client_id': grant.client.client_id,
            'grant_type': grant.grant_type,
            'oidc': bool(response.id_token),
            'refresh_token': bool(response.refresh_token),
            'remote_host': str(self.request.client.host),
            'sector_identifier': grant.client.sector_identifier,
        })
        return response


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import functools
import warnings

from headless.ext.oauth2 import OIDCToken

import cbra.core as cbra
from cbra.core.iam import RequestSubject
from cbra.types import IDependant
from cbra.types import IRequestPrincipal
from cbra.types import ISubject
from cbra.types import ISubjectResolver
from cbra.types import NullRequestPrincipal
from cbra.types import NullSubject
from cbra.types import SessionRequestPrincipal
from cbra.ext.oauth2 import AuthorizationServerStorage


class FrontendSubjectResolver(ISubjectResolver, IDependant):
    __module__: str = 'cbra.core.iam'

    def __init__(
        self,
        storage: AuthorizationServerStorage = cbra.instance('AuthorizationServerStorage')
    ) -> None:
        self.storage = storage

    @functools.singledispatchmethod # type: ignore
    async def resolve(self, principal: IRequestPrincipal) -> ISubject:
        warnings.warn(
            f'{type(self).__name__} does not know how to resolve '
            f'{type(principal).__name__}, returning NullSubject.'
        )
        return NullSubject()

    @resolve.register
    async def resolve_null(
        self,
        principal: NullRequestPrincipal
    ) -> ISubject:
        return NullSubject()

    @resolve.register
    async def _resolve_session(
        self,
        principal: SessionRequestPrincipal
    ) -> ISubject:
        return await self.resolve_session(principal)

    async def resolve_session(
        self,
        principal: SessionRequestPrincipal
    ) -> ISubject:
        assert principal.claims is not None
        subject = NullSubject()
        if principal.claims.sub is not None:
            oidc = await self.storage.get(OIDCToken, principal.subject)
            if oidc is not None:
                subject = RequestSubject(
                    id=oidc.sub,
                    email=oidc.email,
                    principal=principal
                )
        return subject


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi
from headless.ext import oauth2

from cbra.core.conf import settings
from .endpoint import AuthorizationServerEndpoint
from .types import RedirectParameters
from .types import ResponseValidationFailure
from .types import QueryAuthorizeResponse


class AuthorizationCodeCallbackEndpoint(AuthorizationServerEndpoint):
    """Handles a redirect from an OAuth 2.x/OpenID Connect
    authorization server.
    """
    __module__: str = 'cbra.ext.oauth2'
    name: str = 'oauth2.callback'
    nonce_cookie_name: str = 'oidc.nonce'
    redirect_cookie_name: str = 'oauth2.redirect-uri'
    state_cookie_name: str = 'oauth2.state'
    status_code: int = 303
    summary: str = 'Redirection Endpoint'
    tags: list[str] = ['OAuth 2.x/OpenID Connect']

    async def get(
        self,
        client_id: str,
        params: RedirectParameters = RedirectParameters.depends()
    ):
        client = await self.get_client(client_id)
        if client is None:
            return fastapi.responses.PlainTextResponse(
                content=f"The client {client_id} is not configured for this server."
            )
        result = params.result()
        if not isinstance(result, QueryAuthorizeResponse):
            raise NotImplementedError
        if result.state is None:
            return fastapi.responses.PlainTextResponse(
                content="This server requires the 'state' parameter."
            )
        req = await self.storage.get_state(result.state)
        if req is None or not req.is_valid(result):
            return fastapi.responses.PlainTextResponse(content="The request is expired.")
        async with client:
            at = await params.obtain(
                client,
                redirect_uri=(
                    f'{self.request.url.scheme}://'
                    f'{self.request.url.netloc}{self.request.url.path}'
                ),
                state=req.state
            )
            await client.verify_response(at)
            if at.id_token and not await client.verify_oidc(at.id_token, nonce=req.nonce):
                raise ResponseValidationFailure("Invalid OIDC token in response.")
            await self.storage.destroy(req)
            await self._on_success(client, at)
        return fastapi.responses.RedirectResponse(
            status_code=303,
            url=req.redirect_uri or '/'
        )

    async def _on_success(
        self,
        client: oauth2.Client,
        at: oauth2.TokenResponse,
    ) -> None:
        id_token = None
        if at.id_token is not None:
            id_token = oauth2.OIDCToken.parse_jwt(
                token=at.id_token,
                email_verified=client.trust_email
            )
        await self.on_success(client, at.access_token, id_token=id_token)

    async def on_success(
        self,
        client: oauth2.Client,
        access_token: oauth2.BearerTokenCredential,
        id_token: oauth2.OIDCToken | None = None
    ) -> None:
        """Obtained when a valid response with a valid, usable token was
        received from the authorization server.
        """
        raise NotImplementedError

    async def get_client(self, client_id: str) -> oauth2.Client | None:
        """Return a preconfigured OAuth 2.x/OpenID Connect client,
        or ``None`` if the client does not exist.
        """
        # TODO: Quite ugly
        for client in settings.OAUTH2_CLIENTS:
            if client_id in {client.get('name'), client.get('client_id')}:
                instance = oauth2.Client(**client)
                break
        else:
            instance = None
        return instance


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Protocol
from typing import TypeVar

from cbra.core.iam.types import Subject
from .authorizationrequestidentifier import AuthorizationRequestIdentifier
from .iauthorizationflowstate import IAuthorizationFlowState

M = TypeVar('M')


class IAuthorizationServerStorage(Protocol):
    __module__: str = 'cbra.ext.oauth2.types'

    async def get_subject(self, *args: Any, **kwargs: Any) -> Subject | None: ...
    async def fetch(self, oid: Any, cls: type[M]) -> M | None: ...

    async def get(
        self,
        cls: AuthorizationRequestIdentifier | type[M],
        *args: Any,
        **kwargs: Any
    ) -> M | None:
        raise NotImplementedError

    async def persist(
        self,
        obj: Any
    ) -> None:
        ...

    async def destroy(self, obj: Any) -> None: ...
    async def get_state(self, key: str) -> IAuthorizationFlowState | None:
        ...


from typing import TypeAlias
from typing import Union

from canonical import EmailAddress
from headless.ext.oauth2.models import SubjectIdentifier

from .publicidentifier import PublicIdentifier


PrincipalType: TypeAlias = Union[
    EmailAddress,
    SubjectIdentifier,
    PublicIdentifier
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Union

import pydantic

from cbra.core.iam.models import Subject
from .emailscope import EmailScope
from .iauthorizationrequest import IAuthorizationRequest
from .iresourceowner import IResourceOwner
from .openidscope import OIDCScope
from .profilescope import ProfileScope
from .subscope import SubScope
from .useragentfingerprintscope import UserAgentFingerprint


class RequestedScope(pydantic.BaseModel):
    __root__: Union[
        EmailScope,
        OIDCScope,
        ProfileScope,
        SubScope,
        UserAgentFingerprint
    ]

    @property
    def name(self) -> str:
        return self.__root__.name

    @pydantic.root_validator(pre=True)
    def preprocess(cls, values: dict[str, Any]) -> dict[str, Any]:
        root: dict[str, Any] | str | None = values.get('__root__')
        if isinstance(root, str):
            values['__root__'] = root = {'name': root}
        return values

    def apply(
        self,
        subject: Subject,
        owner: IResourceOwner,
        claims: dict[str, Any],
        request: IAuthorizationRequest | None = None
    ) -> None:
        return self.__root__.apply(
            subject=subject,
            owner=owner,
            claims=claims,
            request=request
        )
    
    def requires_consent(self) -> bool:
        return self.__root__.requires_consent()


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

from google.cloud import logging

from cbra.core import Application
from cbra.core.conf import settings
from cbra.core.utils import parent_signature
from .aortadebugendpoint import AortaDebugEndpoint
from .aortaendpoint import AortaEndpoint
from .environ import GOOGLE_DATASTORE_NAMESPACE
from .environ import GOOGLE_HOST_PROJECT
from .environ import GOOGLE_SERVICE_PROJECT


class Service(Application):
    __module__: str = 'cbra.ext.google'

    @parent_signature(Application.__init__)
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        if GOOGLE_SERVICE_PROJECT and GOOGLE_DATASTORE_NAMESPACE:
            from google.cloud.datastore import Client
            self.inject(
                name='GoogleDatastoreClient',
                value=Client(
                    project=GOOGLE_SERVICE_PROJECT,
                    namespace=GOOGLE_DATASTORE_NAMESPACE
                )
            )
            self.container.provide('SubjectResolver', {
                'qualname': 'cbra.ext.google.DatastoreSubjectResolver'
            })
            self.container.provide('SubjectRepository', {
                'qualname': 'cbra.ext.google.DatastoreSubjectRepository'
            })
        self.add(AortaEndpoint, path="/.well-known/aorta")
        if settings.DEBUG:
            self.add(AortaDebugEndpoint, path="/.well-known/aorta/debug")

    def logging_config(self):
        client = logging.Client(project=GOOGLE_HOST_PROJECT or GOOGLE_SERVICE_PROJECT)
        client.setup_logging() # type: ignore
        config = super().logging_config()
        config['formatters']['google-cloud'] = {
            '()': "cbra.ext.google.logging.JSONFormatter",
        }
        config['handlers']['default'] = {
            'class': 'google.cloud.logging.handlers.CloudLoggingHandler',
            'client': client,
            'formatter': 'google-cloud',
            'labels': {
                'kind': 'service'
            }
        }
    
        return config


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import json
from typing import Any
from typing import TypeVar


import pydantic
from google.cloud.datastore import Client as GoogleClient
from headless.ext.oauth2 import OIDCToken
from headless.ext.oauth2.models import SubjectIdentifier

import cbra.core as cbra
from cbra.core.iam.models import Subject
from cbra.ext.google import BaseDatastoreRepository
from cbra.ext.google import DatastoreSubjectRepository
from cbra.ext.google.types import IDatastoreEntity
from cbra.ext.google.types import IDatastoreKey
from cbra.ext.oauth2 import BaseStorage
from cbra.ext.oauth2.models import AuthorizationRequestParameters
from cbra.ext.oauth2.models import Client
from cbra.ext.oauth2.models import ManagedGrant
from cbra.ext.oauth2.models import RefreshToken
from cbra.ext.oauth2.models import ResourceOwner
from cbra.ext.oauth2.types import AuthorizationCode
from cbra.ext.oauth2.types import AuthorizationRequestIdentifier
from cbra.ext.oauth2.types import ResourceOwnerIdentifier
from cbra.ext.oauth2.types import PairwiseIdentifier
from cbra.ext.security import ApplicationKeychain


T = TypeVar('T', bound=pydantic.BaseModel)


class Storage(BaseStorage, BaseDatastoreRepository):
    __module__: str = 'cbra.ext.google.impl.oauth2'
    keychain: ApplicationKeychain
    subjects: DatastoreSubjectRepository

    def __init__(
        self,
        client: GoogleClient | Any = cbra.inject('GoogleDatastoreClient'),
        keychain: ApplicationKeychain = ApplicationKeychain.depends()
    ):
        if not isinstance(client, GoogleClient):
            raise TypeError(f"Invalid client: {repr(client)}")
        self.client = client
        self.keychain = keychain
        self.subjects = DatastoreSubjectRepository(client)

    def entity_to_model(
        self,
        cls: type[T],
        entity: IDatastoreEntity | None,
        **kwargs: Any
    ) -> T | None:
        if entity is None:
            return None
        return cls.parse_obj({**dict(entity), **kwargs}) # type: ignore

    def model_to_entity(
        self,
        key: IDatastoreKey,
        obj: pydantic.BaseModel,
        exclude: set[str] | None = None,
        exclude_none: bool = False
    ) -> IDatastoreEntity:
        # TODO: this is only here to prevent encoding issues with protobuf
        entity = self.entity_factory(key)
        entity.update(json.loads(obj.json(exclude=exclude, exclude_none=exclude_none)))
        return entity

    async def get_authorization_request_by_code(
        self,
        oid: AuthorizationCode
    ) -> AuthorizationRequestParameters | None:
        q = self.query(kind='AuthorizationRequest')
        q.add_filter('code.value', '=', str(oid))
        entity = await self.first(q)
        if entity:
            return self.entity_to_model(
                AuthorizationRequestParameters,
                entity,
                id=entity.key.name
            )

    async def get_authorization_request_by_id(
        self,
        oid: AuthorizationRequestIdentifier
    ) -> AuthorizationRequestParameters | None:
        key = self.key('AuthorizationRequest', str(oid))
        return self.entity_to_model(
            AuthorizationRequestParameters,
            await self.get_entity_by_key(key),
            id=key.name
        )

    async def get_client(self, client_id: str) -> Client | None:
        return None

    async def get_oidc_token(self, oid: SubjectIdentifier) -> OIDCToken | None:
        key = self.key('OIDCToken', oid.sha256)
        return self.entity_to_model(
            OIDCToken,
            await self.get_entity_by_key(key)
        )

    async def get_resource_owner(
        self,
        oid: ResourceOwnerIdentifier
    ) -> ResourceOwner | None:
        key = self.key('ResourceOwner', f'clients/{oid.client_id}/subjects/{oid.sub}')
        return self.entity_to_model(
            ResourceOwner,
            await self.get_entity_by_key(key),
        )

    async def persist_authorization_request(self, obj: AuthorizationRequestParameters) -> None:
        entity = self.model_to_entity(self.key('AuthorizationRequest', obj.id), obj)
        await self.put(entity)

    async def persist_managed_grant(self, obj: ManagedGrant) -> None:
        entity = self.model_to_entity(
            self.key('ManagedGrant', obj.id),
            obj,
            exclude={'id'}
        )
        await self.put(entity)

    async def persist_oidc_token(self, obj: OIDCToken) -> None:
        entity = self.model_to_entity(
            self.key('OIDCToken', obj.subject.sha256),
            obj,
            exclude_none=True
        )
        await self.put(entity)

    async def persist_ppid(self, obj: PairwiseIdentifier) -> None:
        key = self.key(
            'PairwiseIdentifier',
            await self.allocate('PairwiseIdentifier')
        )
        entity = self.entity_factory(key)
        entity.update({'sector': obj.sector, 'sub': obj.sub})
        await self.put(entity)
        assert entity.key.id is not None
        obj.value = entity.key.id

    async def persist_refresh_token(self, obj: RefreshToken) -> None:
        # TODO: Since we havent implement the AuthorizedGrant yet,
        # simply allocate an id.
        if obj.grant_id == 0:
            obj.grant_id = await self.allocate('AuthorizedGrant')
        key = self.key('RefreshToken', str(obj.token))
        await self.put(self.model_to_entity(key, obj, exclude={'token'}))

    async def persist_resource_owner(self, obj: ResourceOwner) -> None:
        key = self.key('ResourceOwner', f'clients/{obj.client_id}/subjects/{obj.sub}')
        await self.put(self.model_to_entity(key, obj))

    async def persist_subject(self, obj: Subject) -> None:
        if not obj.is_encypted():
            await obj.encrypt(self.keychain)
        await self.subjects.persist(obj)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from headless.ext.oauth2.models import ClaimSet

from .endpoint import AuthorizationServerEndpoint


class CurrentSubjectEndpoint(AuthorizationServerEndpoint):
    name: str = 'oauth2.me'
    path: str = '/me'
    summary: str = 'Current User Endpoint'

    async def get(self) -> ClaimSet:
        return ClaimSet(sub='allUsers')\
            if not self.is_authenticated()\
            else ClaimSet(sub='self')


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from datetime import datetime
from datetime import timezone
import functools
from typing import Any

from canonical import EmailAddress
from headless.ext.oauth2.models import OIDCToken
from headless.ext.oauth2.models import SubjectIdentifier

from cbra.core.ioc import instance
from cbra.core.params import CurrentIssuer
from ..memorysubjectrepository import MemorySubjectRepository
from ..models import Subject
from .. import types


class UserOnboardingService:
    """Provides an interface to onboard new and existing subjects."""
    __module__: str = 'cbra.core.iam.services'
    issuer: str
    exclude_claims: set[str] = {'email', 'email_verified', 'sub'}
    subjects: types.ISubjectRepository

    def __init__(
        self,
        issuer: str = CurrentIssuer,
        subjects: types.ISubjectRepository = instance(
            name='SubjectRepository',
            missing=MemorySubjectRepository
        )
    ):
        self.issuer = issuer
        self.subjects = subjects
        self.timestamp = datetime.now(timezone.utc)

    def initialize(self, claims: OIDCToken | None = None) -> Subject:
        return Subject(
            kind='User',
            created=self.timestamp,
            seen=self.timestamp,
            claims=claims.dict(exclude=self.exclude_claims) if claims else {}
        )
    
    async def sync(
        self,
        issuer: str,
        principal: types.PrincipalType
    ) -> tuple[types.Subject, bool]:
        """Add or create a user with the given *immutable* principal."""
        onboarded = False
        subject = await self.subjects.get(principal)
        if not subject:
            onboarded = True
            subject = self.initialize()
            await self.subjects.persist(subject)
            assert subject.uid is not None
            await self.update(subject, issuer, [principal])
        else:
            onboarded = False
        return subject, onboarded

    async def email(
        self,
        issuer: str,
        email: EmailAddress
    ) -> tuple[types.Subject, bool]:
        """Onboard or update a subject using an validated and trusted
        email address.
        """
        raise NotImplementedError

    async def oidc(self, token: OIDCToken) -> tuple[types.Subject, bool]:
        """Onboard or update a subject using an validated and trusted
        OpenID Connect ID Token.
        """
        subject = None
        found = await self.subjects.find_by_principals(token.principals)
        onboarded = False
        if len(found) > 1:
            # The ID token identified multiple subjects and is thus unusable
            # to establish the identity
            raise NotImplementedError(found)
        if not found:
            onboarded = True
            subject = self.initialize(token)
            await self.subjects.persist(subject)
            assert subject.uid is not None
        else:
            assert len(found) == 1
            subject = await self.subjects.get(found.pop())

        # If the subject is None here, then the subject
        # was deleted, but not its principals. This is
        # for now an unrecoverable error.
        if not subject:
            raise NotImplementedError("Missing Subject for Principal(s)")
        assert subject is not None
        principals = token.principals
        if token.email_verified and token.email:
            # Arbitrarily overwrite the email address here if it is trusted,
            # since we might have received it from other sources.
            subject.add_principal(token.iss, token.email, self.timestamp, trust=True)
            subject.seen = self.timestamp
            principals.remove(token.email)
        await self.update_oidc(subject, token)
        return subject, onboarded

    @functools.singledispatchmethod
    async def can_use(
        self,
        obj: OIDCToken | Subject,
        *args: Any,
        **kwargs: Any
    ) -> bool:
        raise NotImplementedError(type(obj).__name__)

    @can_use.register
    async def can_use_oidc_token(
        self,
        token: OIDCToken
    ) -> bool:
        return len(await self.subjects.find_by_principals(token.principals)) <= 1

    @can_use.register
    async def can_use_for_subject(
        self,
        subject: Subject,
        principals: list[types.PrincipalType]
    ) -> bool:
        """Return a boolean indicating if the principals can be used
        by the subject.
        """
        found = await self.subjects.find_by_principals(principals)
        can_use = not found
        if found:
            can_use = (len(found) == 1) and (found.pop() == subject.uid)
        return can_use

    async def destroy(self, uid: int) -> None:
        await self.subjects.destroy(uid)

    async def get(self, oidc: OIDCToken) -> Subject | None:
        found = await self.subjects.find_by_principals(oidc.principals)
        if len(found) > 1:
            raise RuntimeError("Principals resolve to multiple Subjects.")
        if not found:
            return None
        return await self.subjects.get(found.pop()) # type: ignore

    async def update_oidc(
        self,
        subject: types.Subject,
        oidc: OIDCToken
    ) -> None:
        assert subject.uid is not None
        subject.seen = self.timestamp
        if oidc.email and not subject.has_principal(oidc.email):
            subject.add_principal(oidc.iss, oidc.email, self.timestamp, oidc.email_verified)
        identifier = SubjectIdentifier(iss=oidc.iss, sub=oidc.sub)
        if not subject.has_principal(identifier):
            subject.add_principal(oidc.iss, identifier, self.timestamp, True)
        subject.update_oidc(oidc)
        await self.subjects.persist(subject)

    async def update(
        self,
        subject: types.Subject,
        iss: str,
        principals: Any,
        trust: bool = False
    ) -> None:
        assert subject.uid is not None
        subject.seen = self.timestamp
        for principal in principals:
            if subject.has_principal(principal):
                continue
            subject.add_principal(iss, principal, self.timestamp, trust=trust)
        await self.subjects.persist(subject)

    async def persist(self, subject: Subject) -> None:
        await self.subjects.persist(subject)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from headless.ext.oauth2.types import ResponseType

from .accesstype import AccessType
from .authorizationcode import AuthorizationCode
from .authorizationlifecycle import AuthorizationLifecycle
from .authorizationrequestidentifier import AuthorizationRequestIdentifier
from .clientauthenticationmethod import ClientAuthenticationMethod
from .clientinfo import ClientInfo
from .frontendexception import FrontendException
from .grantedscope import GrantedScope
from .iauthorizationserverstorage import IAuthorizationServerStorage
from .fatalauthorizationexception import FatalAuthorizationException
from .fatalclientexception import FatalClientException
from .iauthorizationrequest import IAuthorizationRequest
from .iclient import IClient
from .invalidrequest import InvalidRequest
from .invalidresponsetype import InvalidResponseTypeRequested
from .irefreshtoken import IRefreshToken
from .iresourceowner import IResourceOwner
from .itokenbuilder import ITokenBuilder
from .itokensigner import ITokenSigner
from .jarmauthorizeresponse import JARMAuthorizeResponse
from .queryauthorizeresponse import QueryAuthorizeResponse
from .loginresponse import LoginResponse
from .missingresponsetype import MissingResponseType
from .oidcclaimset import OIDCClaimSet
from .oidcprovider import OIDCProvider
from .pairwiseidentifier import PairwiseIdentifier
from .redirecturi import RedirectURI
from .redirectparameters import RedirectParameters
from .refreshtokenpolicytype import RefreshTokenPolicyType
from .refreshtokentype import RefreshTokenType
from .refreshtokenstatus import RefreshTokenStatus
from .requestedscope import RequestedScope
from .resourceowneridentifier import ResourceOwnerIdentifier
from .responsemodenotsupported import ResponseModeNotSupported
from .responsevalidationfailure import ResponseValidationFailure
from .rfc9068accesstoken import RFC9068AccessToken
from .signableoidctoken import SignableOIDCToken
from .unsupportedauthorizationresponse import UnsupportedAuthorizationResponse
from .usererror import UserError


__all__: list[str] = [
    'AccessType',
    'AuthorizationCode',
    'AuthorizationLifecycle',
    'AuthorizationRequestIdentifier',
    'ClientInfo',
    'ClientAuthenticationMethod',
    'FatalAuthorizationException',
    'FatalClientException',
    'FrontendException',
    'GrantedScope',
    'IAuthorizationRequest',
    'IAuthorizationServerStorage',
    'IClient',
    'InvalidRequest',
    'InvalidResponseTypeRequested',
    'IRefreshToken',
    'IResourceOwner',
    'ITokenBuilder',
    'ITokenSigner',
    'JARMAuthorizeResponse',
    'LoginResponse',
    'MissingResponseType',
    'OIDCClaimSet',
    'OIDCProvider',
    'PairwiseIdentifier',
    'QueryAuthorizeResponse',
    'RedirectURI',
    'RedirectParameters',
    'RefreshTokenPolicyType',
    'RefreshTokenType',
    'RefreshTokenStatus',
    'RequestedScope',
    'ResourceOwnerIdentifier',
    'ResponseModeNotSupported',
    'ResponseType',
    'ResponseValidationFailure',
    'RFC9068AccessToken',
    'SignableOIDCToken',
    'UnsupportedAuthorizationResponse',
    'UserError',
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from cbra.core.iam import SubjectResolver
from cbra.core.iam import RequestSubject
from cbra.core.ioc import instance
from cbra.core.ioc import override
from cbra.types import NullSubject
from cbra.types import SessionRequestPrincipal

from .datastoresubjectrepository import DatastoreSubjectRepository


class DatastoreSubjectResolver(SubjectResolver):
    """A :class:`~cbra.core.iam.SubjectResolver` implementation
    that uses Google Datastore as its storage backend.
    """
    __module__: str = 'cbra.ext.google'
    repo: DatastoreSubjectRepository

    @override(SubjectResolver.__init__)
    def __init__(
        self,
        repo: DatastoreSubjectRepository = instance('SubjectRepository')
    ) -> None:
        self.repo = repo

    async def resolve_session(
        self,
        principal: SessionRequestPrincipal
    ) -> RequestSubject | NullSubject:
        if principal.claims.uid is None:
            return NullSubject()
        subject = await self.repo.resolve(principal.claims.uid)
        if subject is None:
            return NullSubject()
        assert subject.uid is not None # nosec
        return RequestSubject(
            id=str(subject.uid),
            email=principal.claims.email,
            principal=principal
        )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi
from headless.ext.picqer import DefaultClient

from cbra.ext import google
from cbra.ext import picqer



class PicqerWebhookEndpoint(picqer.PicqerWebhookEndpoint):

    async def on_picklists_created(self, event: picqer.v1.PicklistEvent):
        return fastapi.Response(content=event.json(indent=2))


app = google.Service(docs_url='/ui')
app.add(PicqerWebhookEndpoint, path='/ext/picqer.com/v1/.webhooks')


@app.post('/ext/picqer.com/v1/.webhooks/install')
async def f(request: fastapi.Request):
    async with DefaultClient() as client:
        await PicqerWebhookEndpoint.install(
            client,
            callback_url=request.url_for('picqer.webhooks'),
            generate_name=lambda x: (
                f"{str.join('.', reversed(str.split(x, '.')))}"
                f".picqer.{request.url.netloc}"
            ),
            secret='test'
        )


if __name__ == '__main__':
    import uvicorn
    uvicorn.run('__main__:app', reload=True) # type: ignore



# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import base64
import secrets
from datetime import datetime
from datetime import timezone
from typing import Any
from typing import Awaitable
from typing import Callable
from typing import TypeVar

from .isessiondata import ISessionData
from .sessionclaims import SessionClaims
from .sessionmodel import SessionModel


T = TypeVar('T', bound='Session')


class Session(SessionModel, ISessionData[SessionClaims]): # type: ignore

    @classmethod
    def new(cls: type[T]) -> T:
        now = datetime.now(timezone.utc)
        return cls(
            id=secrets.token_urlsafe(48),
            iat=int(now.timestamp())
        )

    def as_cookie(self) -> str:
        assert self.hmac is not None
        v = base64.urlsafe_b64encode(str.encode(self.json(exclude_none=True), encoding='utf-8'))
        return bytes.decode(v, 'ascii')

    def set(self, key: str, value: Any) -> bool:
        if self.claims is None:
            self.claims = SessionClaims()
        if key not in self.claims.__fields__:
            raise AttributeError(f'Claim not supported: {key}')
        modified = False
        if getattr(self.claims, key) != value:
            modified = True
            self.hmac = None
            setattr(self.claims, key, value)
        return modified

    def update(self, claims: dict[str, Any] | SessionClaims) -> None:
        if isinstance(claims, dict):
            claims = SessionClaims.parse_obj(claims)
        if self.claims is None:
            self.claims = SessionClaims()
        self.claims = SessionClaims.parse_obj({
            **self.claims.dict(),
            **claims.dict()
        })
        self.hmac = None

    async def sign(
        self,
        sign: Callable[[bytes], Awaitable[str]]
    ) -> None:
        self.hmac = await sign(self.digest())

    async def verify(
        self,
        verify: Callable[[bytes | str, bytes], Awaitable[bool]]
    ) -> bool:
        assert self.hmac is not None
        return await verify(self.hmac, self.digest())